from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any


class Reference(str, Enum):
    BASE = "base"
    VERSION = "version"


@dataclass(slots=True)
class TransportConfig:
    base_url: str


@dataclass(slots=True)
class ExecutionClientConfig:
    api_key: str
    transport: TransportConfig
    timeout: float = 30.0
    user_agent: str | None = None


@dataclass(slots=True)
class ExecutePolicyRequest:
    id: str
    data: dict[str, Any]
    reference: Reference = Reference.VERSION


@dataclass(slots=True)
class ExecuteFlowRequest:
    id: str
    data: dict[str, Any]
    reference: Reference = Reference.VERSION


@dataclass(slots=True)
class OrchestratorTiming:
    go: str
    database: str
    total: str


@dataclass(slots=True)
class ExecutionTiming:
    orchestrator: OrchestratorTiming | None
    engine: str
    total: str


@dataclass(slots=True)
class PolicyExecutionData:
    result: bool
    trace: Any
    rule: list[str]
    data: Any
    error: Any
    labels: Any


@dataclass(slots=True)
class PolicyExecutionResult:
    kind: str
    result: bool
    trace: Any
    rule: list[str]
    data: Any
    error: Any
    labels: Any
    execution: ExecutionTiming | None = None
    timings: ExecutionTiming | None = None


@dataclass(slots=True)
class FlowNodeExecution:
    database: str
    engine: str
    total: str


@dataclass(slots=True)
class FlowNodeResponse:
    node_id: str
    node_type: str
    response: PolicyExecutionData
    execution: FlowNodeExecution | None = None


@dataclass(slots=True)
class FlowExecutionTiming:
    orchestrator: str
    database: str
    engine: str
    total: str


@dataclass(slots=True)
class FlowExecutionResult:
    kind: str
    result: Any
    node_response: list[FlowNodeResponse] = field(default_factory=list)
    execution: FlowExecutionTiming | None = None
    timings: FlowExecutionTiming | None = None



from .client import ExecutionClient
from .errors import (
    AuthenticationError,
    AuthorizationError,
    ConfigurationError,
    SDKError,
    ServerError,
    TransportError,
)
from .types import (
    ExecuteFlowRequest,
    ExecutePolicyRequest,
    ExecutionClientConfig,
    ExecutionTiming,
    FlowExecutionResult,
    FlowExecutionTiming,
    FlowNodeExecution,
    FlowNodeResponse,
    OrchestratorTiming,
    PolicyExecutionData,
    PolicyExecutionResult,
    Reference,
    TransportConfig,
)

__all__ = [
    "AuthenticationError",
    "AuthorizationError",
    "ConfigurationError",
    "ExecuteFlowRequest",
    "ExecutePolicyRequest",
    "ExecutionClient",
    "ExecutionClientConfig",
    "ExecutionTiming",
    "FlowExecutionResult",
    "FlowExecutionTiming",
    "FlowNodeExecution",
    "FlowNodeResponse",
    "OrchestratorTiming",
    "PolicyExecutionData",
    "PolicyExecutionResult",
    "Reference",
    "SDKError",
    "ServerError",
    "TransportConfig",
    "TransportError",
]



class SDKError(Exception):
    pass


class ConfigurationError(SDKError):
    pass


class TransportError(SDKError):
    pass


class AuthenticationError(SDKError):
    pass


class AuthorizationError(SDKError):
    pass


class ServerError(SDKError):
    def __init__(self, message: str, status: int | None = None):
        super().__init__(message)
        self.status = status



from __future__ import annotations

import io
import json
import unittest
from urllib.error import HTTPError, URLError

from policies2_sdk import (
    AuthenticationError,
    AuthorizationError,
    ConfigurationError,
    ExecutionClient,
    ExecutionClientConfig,
    Reference,
    ServerError,
    TransportConfig,
    TransportError,
)


class FakeResponse:
    def __init__(self, payload: dict):
        self._payload = json.dumps(payload).encode("utf-8")

    def read(self) -> bytes:
        return self._payload

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class ExecutionClientTests(unittest.TestCase):
    def test_requires_api_key(self):
        with self.assertRaises(ConfigurationError):
            ExecutionClient(
                ExecutionClientConfig(api_key="", transport=TransportConfig(base_url="https://example.com"))
            )

    def test_execute_policy_uses_base_path_and_parses_result(self):
        captured = {}

        def opener(request, timeout):
            captured["url"] = request.full_url
            captured["timeout"] = timeout
            captured["api_key"] = request.headers["X-api-key"]
            body = json.loads(request.data.decode("utf-8"))
            captured["body"] = body
            return FakeResponse(
                {
                    "result": True,
                    "trace": {"execution": []},
                    "rule": ["A rule"],
                    "data": {"user": {"age": 25}},
                    "error": None,
                    "labels": None,
                    "execution": {
                        "orchestrator": {"go": "1", "database": "2", "total": "3"},
                        "engine": "4",
                        "total": "5",
                    },
                }
            )

        client = ExecutionClient(
            ExecutionClientConfig(
                api_key="test-key",
                transport=TransportConfig(base_url="https://api.policy2.net/"),
            ),
            opener=opener,
        )

        result = client.execute_policy(
            policy_id="policy-1",
            reference=Reference.BASE,
            data={"user": {"age": 25}},
        )

        self.assertEqual("https://api.policy2.net/run/policy/policy-1", captured["url"])
        self.assertEqual({"data": {"user": {"age": 25}}}, captured["body"])
        self.assertEqual("test-key", captured["api_key"])
        self.assertTrue(result.result)
        self.assertEqual("policy", result.kind)
        self.assertEqual("4", result.execution.engine)

    def test_execute_flow_uses_version_path_and_parses_nodes(self):
        def opener(request, timeout):
            return FakeResponse(
                {
                    "result": {"approved": True},
                    "nodeResponse": [
                        {
                            "nodeId": "node-1",
                            "nodeType": "policy",
                            "response": {
                                "result": True,
                                "trace": None,
                                "rule": ["A rule"],
                                "data": {"ok": True},
                                "error": None,
                                "labels": None,
                            },
                            "execution": {"database": "1", "engine": "2", "total": "3"},
                        }
                    ],
                    "execution": {"orchestrator": "1", "database": "2", "engine": "3", "total": "4"},
                }
            )

        client = ExecutionClient(
            ExecutionClientConfig(api_key="test-key", transport=TransportConfig(base_url="https://api.policy2.net")),
            opener=opener,
        )

        result = client.execute_flow(flow_id="flow-1", data={"user": {"age": 25}})

        self.assertEqual("flow", result.kind)
        self.assertEqual(1, len(result.node_response))
        self.assertEqual("node-1", result.node_response[0].node_id)
        self.assertEqual("3", result.execution.engine)

    def test_maps_401_to_authentication_error(self):
        def opener(request, timeout):
            raise HTTPError(request.full_url, 401, "Unauthorized", hdrs=None, fp=io.BytesIO(b"bad key"))

        client = ExecutionClient(
            ExecutionClientConfig(api_key="bad-key", transport=TransportConfig(base_url="https://api.policy2.net")),
            opener=opener,
        )

        with self.assertRaises(AuthenticationError):
            client.execute_policy(policy_id="policy-1", data={})

    def test_maps_403_to_authorization_error(self):
        def opener(request, timeout):
            raise HTTPError(request.full_url, 403, "Forbidden", hdrs=None, fp=io.BytesIO(b"forbidden"))

        client = ExecutionClient(
            ExecutionClientConfig(api_key="bad-key", transport=TransportConfig(base_url="https://api.policy2.net")),
            opener=opener,
        )

        with self.assertRaises(AuthorizationError):
            client.execute_flow(flow_id="flow-1", data={})

    def test_maps_other_status_to_server_error(self):
        def opener(request, timeout):
            raise HTTPError(request.full_url, 500, "Boom", hdrs=None, fp=io.BytesIO(b"server exploded"))

        client = ExecutionClient(
            ExecutionClientConfig(api_key="key", transport=TransportConfig(base_url="https://api.policy2.net")),
            opener=opener,
        )

        with self.assertRaises(ServerError) as ctx:
            client.execute_policy(policy_id="policy-1", data={})

        self.assertEqual(500, ctx.exception.status)

    def test_maps_url_errors_to_transport_error(self):
        def opener(request, timeout):
            raise URLError("dns down")

        client = ExecutionClient(
            ExecutionClientConfig(api_key="key", transport=TransportConfig(base_url="https://api.policy2.net")),
            opener=opener,
        )

        with self.assertRaises(TransportError):
            client.execute_policy(policy_id="policy-1", data={})

    def test_invalid_json_raises_transport_error(self):
        class BadResponse:
            def read(self) -> bytes:
                return b"{"

            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc, tb):
                return False

        def opener(request, timeout):
            return BadResponse()

        client = ExecutionClient(
            ExecutionClientConfig(api_key="key", transport=TransportConfig(base_url="https://api.policy2.net")),
            opener=opener,
        )

        with self.assertRaises(TransportError):
            client.execute_policy(policy_id="policy-1", data={})


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
from typing import Any, Callable
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from .errors import (
    AuthenticationError,
    AuthorizationError,
    ConfigurationError,
    ServerError,
    TransportError,
)
from .types import (
    ExecuteFlowRequest,
    ExecutePolicyRequest,
    ExecutionClientConfig,
    ExecutionTiming,
    FlowExecutionResult,
    FlowExecutionTiming,
    FlowNodeExecution,
    FlowNodeResponse,
    OrchestratorTiming,
    PolicyExecutionData,
    PolicyExecutionResult,
    Reference,
)


class ExecutionClient:
    def __init__(
        self,
        config: ExecutionClientConfig,
        opener: Callable[..., Any] = urlopen,
    ) -> None:
        if not config.api_key.strip():
            raise ConfigurationError("api_key is required")
        if not config.transport.base_url.strip():
            raise ConfigurationError("transport.base_url is required")

        self._api_key = config.api_key
        self._base_url = config.transport.base_url.rstrip("/")
        self._timeout = config.timeout
        self._user_agent = config.user_agent
        self._opener = opener

    def execute_policy(
        self,
        request: ExecutePolicyRequest | None = None,
        *,
        policy_id: str | None = None,
        data: dict[str, Any] | None = None,
        reference: Reference = Reference.VERSION,
    ) -> PolicyExecutionResult:
        req = request or ExecutePolicyRequest(id=policy_id or "", data=data or {}, reference=reference)
        payload = self._send(self._policy_path(req.id, req.reference), req.data)
        return self._parse_policy_result(payload, kind="policy")

    def execute_flow(
        self,
        request: ExecuteFlowRequest | None = None,
        *,
        flow_id: str | None = None,
        data: dict[str, Any] | None = None,
        reference: Reference = Reference.VERSION,
    ) -> FlowExecutionResult:
        req = request or ExecuteFlowRequest(id=flow_id or "", data=data or {}, reference=reference)
        payload = self._send(self._flow_path(req.id, req.reference), req.data)
        return self._parse_flow_result(payload, kind="flow")

    def _send(self, path: str, data: dict[str, Any]) -> dict[str, Any]:
        body = json.dumps({"data": data}).encode("utf-8")
        headers = {
            "content-type": "application/json",
            "x-api-key": self._api_key,
        }
        if self._user_agent:
            headers["user-agent"] = self._user_agent

        request = Request(
            url=f"{self._base_url}{path}",
            data=body,
            headers=headers,
            method="POST",
        )

        try:
            with self._opener(request, timeout=self._timeout) as response:
                payload = response.read().decode("utf-8")
        except HTTPError as exc:
            try:
                payload = exc.read().decode("utf-8").strip()
            finally:
                exc.close()
            if exc.code == 401:
                raise AuthenticationError(payload or "request rejected: invalid API key") from exc
            if exc.code == 403:
                raise AuthorizationError(payload or "request rejected: insufficient permissions") from exc
            raise ServerError(payload or f"request failed with status {exc.code}", status=exc.code) from exc
        except URLError as exc:
            raise TransportError("REST execution request failed") from exc

        try:
            return json.loads(payload)
        except json.JSONDecodeError as exc:
            raise TransportError("failed to decode response body") from exc

    @staticmethod
    def _policy_path(identifier: str, reference: Reference) -> str:
        return f"/run/policy/{identifier}" if reference == Reference.BASE else f"/run/policy_version/{identifier}"

    @staticmethod
    def _flow_path(identifier: str, reference: Reference) -> str:
        return f"/run/flow/{identifier}" if reference == Reference.BASE else f"/run/flow_version/{identifier}"

    @staticmethod
    def _parse_execution_timing(value: Any) -> ExecutionTiming | None:
        if not isinstance(value, dict):
            return None
        orchestrator = value.get("orchestrator")
        parsed_orchestrator = None
        if isinstance(orchestrator, dict):
            parsed_orchestrator = OrchestratorTiming(
                go=str(orchestrator.get("go", "")),
                database=str(orchestrator.get("database", "")),
                total=str(orchestrator.get("total", "")),
            )
        return ExecutionTiming(
            orchestrator=parsed_orchestrator,
            engine=str(value.get("engine", "")),
            total=str(value.get("total", "")),
        )

    @staticmethod
    def _parse_flow_execution_timing(value: Any) -> FlowExecutionTiming | None:
        if not isinstance(value, dict):
            return None
        return FlowExecutionTiming(
            orchestrator=str(value.get("orchestrator", "")),
            database=str(value.get("database", "")),
            engine=str(value.get("engine", "")),
            total=str(value.get("total", "")),
        )

    def _parse_policy_result(self, payload: dict[str, Any], *, kind: str) -> PolicyExecutionResult:
        return PolicyExecutionResult(
            kind=kind,
            result=bool(payload.get("result", False)),
            trace=payload.get("trace"),
            rule=list(payload.get("rule", [])),
            data=payload.get("data"),
            error=payload.get("error", payload.get("errors")),
            labels=payload.get("labels"),
            execution=self._parse_execution_timing(payload.get("execution")),
            timings=self._parse_execution_timing(payload.get("timings")),
        )

    def _parse_flow_result(self, payload: dict[str, Any], *, kind: str) -> FlowExecutionResult:
        node_response = []
        for node in payload.get("nodeResponse", []):
            response = node.get("response", {})
            parsed_execution = node.get("execution")
            node_response.append(
                FlowNodeResponse(
                    node_id=str(node.get("nodeId", "")),
                    node_type=str(node.get("nodeType", "")),
                    response=PolicyExecutionData(
                        result=bool(response.get("result", False)),
                        trace=response.get("trace"),
                        rule=list(response.get("rule", [])),
                        data=response.get("data"),
                        error=response.get("error", response.get("errors")),
                        labels=response.get("labels"),
                    ),
                    execution=(
                        FlowNodeExecution(
                            database=str(parsed_execution.get("database", "")),
                            engine=str(parsed_execution.get("engine", "")),
                            total=str(parsed_execution.get("total", "")),
                        )
                        if isinstance(parsed_execution, dict)
                        else None
                    ),
                )
            )

        return FlowExecutionResult(
            kind=kind,
            result=payload.get("result"),
            node_response=node_response,
            execution=self._parse_flow_execution_timing(payload.get("execution")),
            timings=self._parse_flow_execution_timing(payload.get("timings")),
        )
